// Copyright 2000-2005 the Contributors, as shown in the revision logs.
// Licensed under the Apache Public Source License 2.0 ("the License").
// You may not use this file except in compliance with the License.

package org.ibex.util;

/** A HashMap with a fixed size; drops extraneous elements. Uses an
 *  LRU strategy, either ordering elements based on insertion order
 *  or access order.
 *
 *  @author crawshaw@ibex.org
 */
public class Cache extends Basket.HashMap {
    private static final long serialVersionUID = 23498092L;

    private final int maxSize;
    private final boolean accessOrder;

    private int[] prev;
    private int[] next;
    private int first = -1, last = -1;

    public Cache(int maxSize, boolean accessOrder) {
        super(maxSize * 2, 0.75F);

        prev = new int[maxSize];
        next = new int[maxSize];
        for (int i=0; i < maxSize; i++) { prev[i] = next[i] = -1; }

        this.maxSize = maxSize;
        this.accessOrder = accessOrder;
    }

    public Object get(Object k) {
        int i = indexOf(k);
        if (i < 0) return null;
        if (accessOrder) { // move to front of linked list
            entryRemoved(i);
            prev[i] = -1;
            if (first >= 0) prev[first] = i;
            next[i] = first;
            first = i;
        }
        return entries[i + 1];
    }

    protected void entryAdded(int i) {
        if (last >= 0) next[last] = i;
        prev[i] = last >= 0 ? last : -1;
        next[i] = -1;
        int ourlast = accessOrder ? last : first;
        last = i;
        if (first < 0) first = last;
        if (ourlast >= 0 && size > maxSize) remove(ourlast);
    }

    protected void entryRemoved(int i) {
        if (prev[i] >= 0) next[prev[i]] = next[i];
        if (next[i] >= 0) prev[next[i]] = prev[i];
        if (last == i) last = prev[i] >= 0 ? prev[i] : -1;
        if (first == i) first = next[i] >= 0 ? next[i] : -1;
    }

    protected void entryUpdated(int i) {
        if (!accessOrder) { entryRemoved(i); entryAdded(i); }
    }
}
